#include <map.h>
#include <stdlib.h>

Map map_create(uint32_t buckets_count, uint32_t initial_bucket_size) {
  Map zero = {};

  Entry** buckets = malloc(sizeof(Entry*)*buckets_count);
  if(buckets == 0) {
    return zero;
  }

  uint32_t* bucket_sizes = malloc(sizeof(uint32_t)*buckets_count);
  if(bucket_sizes == 0) {
    free(buckets);
    return zero;
  }

  uint32_t* bucket_usage = malloc(sizeof(uint32_t)*buckets_count);
  if(bucket_usage == 0) {
    free(bucket_usage);
    free(buckets);
    return zero;
  }

  for(uint32_t i = 0; i < buckets_count; i++) {
    Entry* bucket = malloc(sizeof(Entry)*initial_bucket_size);
    if(bucket == 0) {
      for(uint32_t j = 0; j < i; j++) {
        free(buckets[j]);
      }
      free(bucket_sizes);
      free(buckets);
      return zero;
    }
    buckets[i] = bucket;
    bucket_sizes[i] = initial_bucket_size;
    bucket_usage[i] = 0;
  }

  Map ret = {
    .buckets_count = buckets_count,
    .bucket_sizes = bucket_sizes,
    .bucket_usage = bucket_usage,
    .buckets = buckets,
  };

  return ret;
}

MaybeValue map_lookup(Map map, uint32_t key) {
  MaybeValue ret = {
    .has_value = false,
    .value = 0,
  };

  uint32_t bucket_index = key % map.buckets_count;
  for(uint32_t i = 0; i < map.bucket_usage[bucket_index]; i++) {
    if(map.buckets[bucket_index][i].key == key) {
      ret.has_value = true;
      ret.value = map.buckets[bucket_index][i].value;
    }
  }

  return ret;
}

bool map_add(Map* map, uint32_t key, void* value) {
  uint32_t bucket_index = key % map->buckets_count;

  for(uint32_t i = 0; i < map->bucket_usage[bucket_index]; i++) {
    if(map->buckets[bucket_index][i].key == key) {
      map->buckets[bucket_index][i].value = value;
      return true;
    }
  }

  if(map->bucket_usage[bucket_index] < map->bucket_sizes[bucket_index]) {
    map->buckets[bucket_index][map->bucket_usage[bucket_index]].key = key;
    map->buckets[bucket_index][map->bucket_usage[bucket_index]].value = value;
    map->bucket_usage[bucket_index] += 1;
    return true;
  }

  Entry* new_bucket = realloc(map->buckets[bucket_index], 2*map->bucket_sizes[bucket_index]);
  if(new_bucket == 0) {
    return false;
  }

  map->bucket_usage[bucket_index] += 1;
  map->bucket_sizes[bucket_index] *= 2;
  map->buckets[bucket_index] = new_bucket;
  map->buckets[bucket_index][map->bucket_usage[bucket_index]].key = key;
  map->buckets[bucket_index][map->bucket_usage[bucket_index]].value = value;

  return true;
}

MaybeValue map_del(Map* map, uint32_t key) {
  MaybeValue ret = {
    .has_value = false,
    .value = 0,
  };

  uint32_t bucket_index = key % map->buckets_count;
  for(uint32_t i = 0; i < map->bucket_usage[bucket_index]; i++) {
    if(map->buckets[bucket_index][i].key == key) {
      ret.value = map->buckets[bucket_index][i].value;
      ret.has_value = true;

      if(map->bucket_usage[bucket_index] > 1) {
        map->buckets[bucket_index][i] = map->buckets[bucket_index][map->bucket_usage[bucket_index]-1];
      }
      map->bucket_usage[bucket_index] -= 1;
      break;
    }
  }

  return ret;
}

void map_destroy(Map map) {
  for(uint32_t i = 0; i < map.buckets_count; i++) {
    free(map.buckets[i]);
  }
  free(map.buckets);
  free(map.bucket_sizes);
  free(map.bucket_usage);
}

MapIterator map_iterator_create(Map map) {
  MapIterator iterator = {
    .keys  = 0,
    .vals  = 0,
    .count = 0,
  };

  uint32_t count = 0;
  for(uint32_t i = 0; i < map.buckets_count; i++) {
    count += map.bucket_usage[i];
  }

  uint32_t* keys = malloc(sizeof(uint32_t)*count);
  if(keys == 0) {
    return iterator;
  }

  void** vals = malloc(sizeof(void*)*count);
  if(vals == 0) {
    free(keys);
    return iterator;
  }

  uint32_t idx = 0;
  for(uint32_t i = 0; i < map.buckets_count; i++) {
    for(uint32_t j = 0; j < map.bucket_usage[i]; j++) {
      keys[idx] = map.buckets[i][j].key;
      vals[idx] = map.buckets[i][j].value;
      idx += 1;
    }
  }

  iterator.keys = keys;
  iterator.vals = vals;
  iterator.count = count;

  return iterator;
}

void map_iterator_free(MapIterator iterator) {
  if(iterator.count > 0) {
    free(iterator.keys);
    free(iterator.vals);
  }
}
